import pandas as pd
from pandas.testing import assert_frame_equal

from upgini.metadata import SearchKey
from upgini.utils.email_utils import EmailSearchKeyConverter, EmailSearchKeyDetector

detector = EmailSearchKeyDetector()


def test_email_search_key_detection_by_name():
    df = pd.DataFrame({"email": ["123", "321", "555"]})

    assert detector.get_search_key_column(df) == "email"

    df = pd.DataFrame({"e-mail": ["123", "321", "555"]})

    assert detector.get_search_key_column(df) == "e-mail"

    df = pd.DataFrame({"e_mail": ["123", "321", "555"]})

    assert detector.get_search_key_column(df) == "e_mail"


def test_email_search_key_detection_by_values():
    df = pd.DataFrame({"eml": ["asdf@asdf.sad", "woei@skdjfh.fnj"] + ["12@3"] * 8})

    assert detector.get_search_key_column(df) == "eml"

    df = pd.DataFrame({"eml": ["asdf@asdf.sad"] + ["12@"] * 9})

    maybe_search_key_column = detector.get_search_key_column(df)
    print(maybe_search_key_column)
    assert maybe_search_key_column is None

    df = pd.DataFrame({"eml": [1, 2, 3, 4, 5]})

    assert detector.get_search_key_column(df) is None


def test_convertion_to_hem():
    df = pd.DataFrame({"email": ["test@google.com", "", "@", None, 0.0, "asdf@oiouo@asdf"]})

    search_keys = {"email": SearchKey.EMAIL}
    converter = EmailSearchKeyConverter("email", None, search_keys)
    df = converter.convert(df)

    assert search_keys == {
        EmailSearchKeyConverter.HEM_COLUMN_NAME: SearchKey.HEM,
        EmailSearchKeyConverter.EMAIL_ONE_DOMAIN_COLUMN_NAME: SearchKey.EMAIL_ONE_DOMAIN,
    }

    expected_df = pd.DataFrame(
        {
            "email": ["test@google.com", "", "@", None, 0.0, "asdf@oiouo@asdf"],
            EmailSearchKeyConverter.HEM_COLUMN_NAME: [
                "8b0080a904da73e6e500ada3d09a88037289b5c08e03d3a09546ffacc5b5fd57",
                None,
                None,
                None,
                None,
                None,
            ],
            EmailSearchKeyConverter.EMAIL_ONE_DOMAIN_COLUMN_NAME: ["tgoogle.com", None, None, None, None, None],
            EmailSearchKeyConverter.DOMAIN_COLUMN_NAME: ["google.com", None, None, None, None, None],
        }
    )

    assert_frame_equal(expected_df, df)

    assert df[EmailSearchKeyConverter.DOMAIN_COLUMN_NAME].astype("string").isnull().sum() == 5


def test_convertion_to_hem_with_existing_hem():
    df = pd.DataFrame(
        {
            "email": ["test@google.com", "", None, 0.0],
            "hem": ["8b0080a904da73e6e500ada3d09a88037289b5c08e03d3a09546ffacc5b5fd57", None, None, None],
        }
    )

    search_keys = {"email": SearchKey.EMAIL, "hem": SearchKey.HEM}
    converter = EmailSearchKeyConverter("email", "hem", search_keys)
    df = converter.convert(df)

    assert search_keys == {
        "hem": SearchKey.HEM,
        EmailSearchKeyConverter.EMAIL_ONE_DOMAIN_COLUMN_NAME: SearchKey.EMAIL_ONE_DOMAIN,
    }

    expected_df = pd.DataFrame(
        {
            "email": ["test@google.com", "", None, 0.0],
            "hem": ["8b0080a904da73e6e500ada3d09a88037289b5c08e03d3a09546ffacc5b5fd57", None, None, None],
            EmailSearchKeyConverter.EMAIL_ONE_DOMAIN_COLUMN_NAME: ["tgoogle.com", None, None, None],
            EmailSearchKeyConverter.DOMAIN_COLUMN_NAME: ["google.com", None, None, None],
        }
    )

    assert_frame_equal(expected_df, df)



import pandas as pd

from pandas.testing import assert_frame_equal

from upgini.utils.country_utils import CountrySearchKeyDetector

detector = CountrySearchKeyDetector()


def test_is_country_column_by_values():
    df = pd.DataFrame({"cntr": ["not country"] * 8 + ["ES", "IT"]})

    assert detector.get_search_key_column(df) == "cntr"

    df = pd.DataFrame({"cntr": ["not country"] * 9 + ["IT"]})

    assert detector.get_search_key_column(df) is None

    df = pd.DataFrame({"cntr": [1, 2, 3, 4, 5]})

    assert detector.get_search_key_column(df) is None


def test_is_country_column_by_name():
    df = pd.DataFrame({"country": ["not country"] * 10})

    assert detector.get_search_key_column(df) == "country"


def test_country_to_iso_code_convertion():
    df = pd.DataFrame(
        {
            "country": [
                "Austria",
                "England",
                "Poland",
                "United Kingdom of Great Britain and Northern Ireland",
                "",
                "Unknown",
                "US",
                "United States",
            ]
        }
    )

    expected_df = df.copy()
    expected_df["country"] = ["AT", "GB", "PL", "GB", "", "Unknown", "US", "US"]

    df_with_code = CountrySearchKeyDetector.convert_country_to_iso_code(df, "country")

    assert_frame_equal(expected_df, df_with_code)



import abc
from typing import Any, List, Optional, Union
import numpy as np
import pandas as pd
from pydantic import BaseModel

from upgini.autofe.operand import PandasOperand


class DateDiffMixin(BaseModel):
    diff_unit: str = "D"
    left_unit: Optional[str] = None
    right_unit: Optional[str] = None

    def _convert_to_date(
        self, x: Union[pd.DataFrame, pd.Series], unit: Optional[str]
    ) -> Union[pd.DataFrame, pd.Series]:
        if isinstance(x, pd.DataFrame):
            return x.apply(lambda y: self._convert_to_date(y, unit), axis=1)

        return pd.to_datetime(x, unit=unit)


class DateDiff(PandasOperand, DateDiffMixin):
    name = "date_diff"
    is_binary = True
    has_symmetry_importance = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        left = self._convert_to_date(left, self.left_unit)
        right = self._convert_to_date(right, self.right_unit)
        return self.__replace_negative((left - right) / np.timedelta64(1, self.diff_unit))

    def __replace_negative(self, x: Union[pd.DataFrame, pd.Series]):
        x[x < 0] = None
        return x


class DateDiffType2(PandasOperand, DateDiffMixin):
    name = "date_diff_type2"
    is_binary = True
    has_symmetry_importance = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        left = self._convert_to_date(left, self.left_unit)
        right = self._convert_to_date(right, self.right_unit)
        future = right + (left.dt.year - right.dt.year).apply(
            lambda y: np.datetime64("NaT") if np.isnan(y) else pd.tseries.offsets.DateOffset(years=y)
        )
        before = future[future < left]
        future[future < left] = before + pd.tseries.offsets.DateOffset(years=1)
        diff = (future - left) / np.timedelta64(1, self.diff_unit)

        return diff


class DateListDiff(PandasOperand, DateDiffMixin):
    is_binary = True
    has_symmetry_importance = True
    aggregation: str

    def __init__(self, **data: Any) -> None:
        if "name" not in data:
            data["name"] = f"date_diff_{data.get('aggregation')}"
        super().__init__(**data)

    def map_diff(self, left: np.datetime64, right: list) -> list:
        return (left - self._convert_to_date(pd.Series(right), self.right_unit)) / np.timedelta64(1, self.diff_unit)

    def reduce(self, diff_list: pd.Series) -> float:
        return diff_list[diff_list > 0].aggregate(self.aggregation)

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        left = self._convert_to_date(left, self.left_unit)

        return pd.Series(left.index.map(lambda i: self.reduce(self.map_diff(left.loc[i], right.loc[i]))))


class DateListDiffBounded(DateListDiff):
    lower_bound: Optional[int]
    upper_bound: Optional[int]
    inclusive: Optional[str]

    def __init__(self, **data: Any) -> None:
        if "name" not in data:
            inclusive = data.get("inclusive")
            lower_bound = data.get("lower_bound")
            upper_bound = data.get("upper_bound")
            components = [
                "date_diff",
                data.get("diff_unit"),
                str(lower_bound if lower_bound is not None else "minusinf"),
                str(upper_bound if upper_bound is not None else "plusinf"),
            ]
            if inclusive:
                components.append(inclusive)
            components.append(data.get("aggregation"))
            data["name"] = "_".join(components)
        super().__init__(**data)

    def reduce(self, diff_list: pd.Series) -> float:
        return diff_list[
            (diff_list > 0)
            & (
                diff_list.between(
                    self.lower_bound or -np.inf, self.upper_bound or np.inf, inclusive=self.inclusive or "left"
                )
            )
        ].aggregate(self.aggregation)



from typing import Dict
from upgini.autofe.date import DateDiff, DateDiffType2
from upgini.autofe.groupby import GroupByThenAgg, GroupByThenRank
from upgini.autofe.operand import Operand
from upgini.autofe.unary import Abs, Log, Residual, Sqrt, Square, Sigmoid, Floor, Freq
from upgini.autofe.binary import Min, Max, Add, Subtract, Multiply, Divide, Sim
from upgini.autofe.vector import Mean, Sum

ALL_OPERANDS: Dict[str, Operand] = {
    op.name: op
    for op in [
        Freq(),
        Mean(),
        Sum(),
        Abs(),
        Log(),
        Sqrt(),
        Square(),
        Sigmoid(),
        Floor(),
        Residual(),
        Min(),
        Max(),
        Add(),
        Subtract(),
        Multiply(),
        Divide(),
        GroupByThenAgg(name="GroupByThenMin", agg="min"),
        GroupByThenAgg(name="GroupByThenMax", agg="max"),
        GroupByThenAgg(name="GroupByThenMean", agg="mean"),
        GroupByThenAgg(name="GroupByThenMedian", agg="median"),
        GroupByThenAgg(name="GroupByThenStd", output_type="float", agg="std"),
        GroupByThenRank(),
        Operand(name="Combine", has_symmetry_importance=True, output_type="object", is_categorical=True),
        Operand(name="CombineThenFreq", has_symmetry_importance=True, output_type="float"),
        Operand(name="GroupByThenNUnique", output_type="int", is_vectorizable=True, is_grouping=True),
        Operand(name="GroupByThenFreq", output_type="float", is_grouping=True),
        Sim(),
        DateDiff(),
        DateDiffType2(),
    ]
}


def find_op(name):
    return ALL_OPERANDS.get(name)
